########################################################################
# 2025 April 14
#
# The author disclaims copyright to this source code.  In place of
# a legal notice, here is a blessing:
#
#  * May you do good and not evil.
#  * May you find forgiveness for yourself and forgive others.
#  * May you share freely, never taking more than you give.
#
########################################################################

# @teaish-check-ncurses
#
# Curses!
#
# Jumps through numerous hoops to try to find ncurses libraries and
# appropriate compilation/linker flags. Returns 0 on failure, 1 on
# success. Updates the teaish CFLAGS and LDFLAGS appropriately.
#
# This impl prefers to use pkg-config to find ncurses and libpanel
# because various platforms either combine, or not, the wide-char
# versions of those libs into the same library (or not). If no
# pkg-config is available, OR the platform looks like Mac, then we
# simply make an educated guess and hope it works. On Mac pkg-config
# is not sufficient because the core system and either brew or
# macports can contain mismatched versions of ncurses and iconv, so on
# that platform we simply guess from the core system level, ignoring
# brew/macports options.
#
# PS: these checks could be improved considerably, but they "work for
# me" as-is.
proc teaish-check-ncurses {} {
  use pkg-config
  teaish-check-cached -nomsg -nostatus "Looking for ncurses" {
    # ----------------^^^^^^ the sub-tests will emit all we need. We
    # could instead use -quiet but when this test fails, it's useful
    # to be able to see which parts of it passed/failed so that this
    # check can be improved.
    set LIB_CURSES ""
    set LDFLAGS_CURSES ""
    set CFLAGS_CURSES ""
    set rc 0
    if {1 &&
        $::tcl_platform(os) ne "Darwin"
        && [msg-quiet pkg-config-init 0]} {
      # Some macOS pkg-config configurations alter library search paths, which make
      # the compiler unable to find lib iconv, so don't use pkg-config on macOS.
      set np ""
      foreach p {ncursesw ncurses} {
        if {[pkg-config $p]} {
          set np $p
          msg-result "Using pkg-config curses package \[$p]"
          break
        }
      }
      if {"" ne $np} {
        set ppanel ""
        if {"ncursesw" eq $np} {
          if {[pkg-config panelw]} {
            set ppanel panelw
          }
        }
        if {"" eq $ppanel && [pkg-config panel]} {
          set ppanel panel
        }
        set P [string toupper $ppanel]
        set N [string toupper $np]
        set CFLAGS_CURSES \
          [join [list [get-define PKG_${N}_CFLAGS ""] [get-define PKG_${P}_CFLAGS ""]]]
        set LIB_CURSES [get-define PKG_${N}_LIBS ""]
        set LDFLAGS_CURSES [get-define PKG_${N}_LDFLAGS ""]
        if {"" eq $ppanel} {
          # Apparently Mac brew has pkg-config for ncursesw but not
          # panel/panelw, but hard-coding -lpanel seems to work on
          # that platform.
          append LIB_CURSES " -lpanel"
        } else {
          append LIB_CURSES " " [get-define PKG_${P}_LIBS ""]
          append LDFLAGS_CURSES [get-define PKG_${P}_LDFLAGS ""]
          # append CFLAGS_CURSES " " [exec $pcBin --cflags $ppanel]
          # ^^^^ appending the panel cflags will end up duplicating
          # at least one -D flag from $np's cflags, leading to
          # "already defined" errors at compile-time. Sigh. Note, however,
          # that $ppanel's cflags have flags which $np's do not, so we
          # may need to include those flags anyway and manually perform
          # surgery on the list to remove dupes. Sigh.
        }
      }
    }

    if {"" eq $LIB_CURSES} {
      #puts "Guessing curses location (will fail for exotic locations)..."
      if {[cc-check-includes curses.h]} {
        # Linux has -lncurses, BSD -lcurses. Both have <curses.h>
        msg-result "Found curses.h"
        if {[proj-check-function-in-lib waddnwstr ncursesw]} {
          msg-result "Found -lncursesw"
          set LIB_CURSES "-lncursesw -lpanelw"
        } elseif {[proj-check-function-in-lib initscr ncurses]} {
          msg-result "Found -lncurses"
          set LIB_CURSES "-lncurses -lpanel"
        } elseif {[proj-check-function-in-lib initscr curses]} {
          msg-result "Found -lcurses"
          set LIB_CURSES "-lcurses -lpanel"
        }
      }
    }
    if {"" ne $LIB_CURSES} {
      incr rc
      proj-indented-notice -notice {
        ************************************************************
        If your build of fails due to missing ncurses functions
        such as waddwstr(), make sure you have the ncursesw (with a
        "w") development package installed. Some platforms combine
        the "w" and non-w curses builds and some don't. Similarly,
        it's easy to get a mismatch between libncursesw and libpanel,
        so make sure you have libpanelw (if appropriate for your
        platform).

        The package may have a name such as libncursesw5-dev or
        some such.
        ************************************************************
      }
    }
    if {[define HAVE_CURSES $rc]} {
      teaish-ldflags-prepend $LDFLAGS_CURSES $LIB_CURSES
      teaish-cflags-add $CFLAGS_CURSES
    }
    return $rc
  }
}; # teaish-check-ncurses
