########################################################################
# 2025 May 1
#
# The author disclaims copyright to this source code.  In place of
# a legal notice, here is a blessing:
#
#  * May you do good and not evil.
#  * May you find forgiveness for yourself and forgive others.
#  * May you share freely, never taking more than you give.
#
########################################################################
#
# Generic helpers commonly helpful in/around the wanderinghorse.net
# website.
#
#package provide wh 0.1
namespace eval wh {
  set tclDir [file dirname [file normalize [info script]]]
  #if {$tclDir ni $::auto_path} {
  #  set ::auto_path [linsert $::auto_path 0 $tclDir]
  #}

  variable libdir $tclDir;  # dir this file lives in
  variable argvFlags;       # --flag entries in $::argv
  variable argvNonFlags;    # non-flag entries in $::argv

  unset tclDir

  # Internal config state
  variable config
  array set config {
    verbose-affirm 0
    verbose-assert 0
  }

  # Internal counters
  variable counters
  array set counters {}

  #
  # Returns the name of the scope $lvl-1 levels above this
  # call in the state. So pass 0 for the calling scope's name,
  # 1 for it's calling scope's name, etc.
  #
  proc scope {{lvl 0}} {
    set ilvl [info level]
    set offset [expr {$ilvl  - $lvl - 1}]
    if { $offset < 0} {
      return "invalid scope ($offset)"
    } elseif { $offset == 0} {
      return "global scope"
    } else {
      return [lindex [info level $offset] 0]
    }
  }

  #
  # @warn msg
  #
  # Emits a warning message to stderr. All args are appended with a
  # space between each.
  #
  proc warn {args} {
    puts stderr [join [list "WARNING: \[[scope 1]\]: " {*}$args] " "]
  }

  # Internal impl of [fatal] and [error]
  proc __faterr {failMode argv} {
    set lvl 2
    while {"-up" eq [lindex $argv 0]} {
      set argv [lassign $argv -]
      incr lvl
    }
    if {$failMode} {
      puts stderr [join [list "FATAL: \[[scope $lvl]]: " {*}$argv]]
      exit 1
    } else {
      error [join [list "\[[scope $lvl]]:" {*}$argv]]
    }
  }

  #
  # @fatal ?-up...? msg...
  #
  # Emits an error message to stderr and exits with non-0. All args are
  # appended with a space between each.
  #
  # The calling scope's name is used in the error message. To instead
  # use the name of a call higher up in the stack, use -up once for each
  # additional level.
  #
  proc fatal {args} {
    __faterr 1 $args
  }

  #
  # @error ?-up...? msg...
  #
  # Works like [fatal] but uses [error] intead of [exit].
  #
  proc error {args} {
    __faterr 0 $args
  }

  # Internal impl for affirm and assert.
  #
  # $args = ?-v? script {msg ""}
  proc __affert {failMode args} {
    variable config;
    if {$failMode} {
      set what assert
    } else {
      set what affirm
    }
    set verbose $config(verbose-$what)
    if {"-v" eq [lindex $args 0]} {
      lassign $args - script msg
      if {1 == [llength $args]} {
        # If -v is the only arg, toggle default verbose mode
        set config(verbose-$what) [expr {!$config(verbose-$what)}]
        return
      }
      incr verbose
    } else {
      lassign $args script msg
    }
    variable counters;
    incr counters($what)
    if {![uplevel 1 [concat expr [list $script]]]} {
      if {"" eq $msg} {
        set msg $script
      }
      set txt [join [list $what # $counters($what) "failed:" $script]]
      if {$failMode} {
        puts stderr $txt
        exit 1
      } else {
        error $txt
      }
    } elseif {$verbose} {
      puts stderr [join [list $what # $counters($what) "passed:" $script]]
    }
  }

  #
  # Usages:
  #
  #  affirm ?-v? script ?msg?
  #
  # Works like a conventional assert method does, but reports failures
  # using [error] instead of [exit]. If -v is used, it reports passing
  # assertions to stderr. $script is evaluated in the caller's scope as
  # an argument to [expr].
  #
  proc affirm {args} {
    tailcall __affert 0 {*}$args
  }

  #
  # Works like [affirm] but exits on error.
  #
  proc assert {args} {
    tailcall __affert 1 {*}$args
  }

  #
  # Returns a "unique" name with the given prefix each time it is
  # called in a given runtime session. Barring integer overloads, of
  # course.
  #
  proc unique-name {{prefix "unique-name"}} {
    variable counters;
    return $prefix#[incr counters(unique-name:$prefix)]
  }

  #
  # Expects to be passed one or more flags which might be in
  # $::argv. If found in $::argv, the first instance of the flag is
  # removed from $::argv. Returns the total number of entries removed.
  #
  proc remove-arg {args} {
    set rv 0
    foreach flag $args {
      set n [lsearch $::argv $flag]
      if {$n > -1} {
        set ::argv [lreplace $::argv $n $n]
        incr rv
      }
    }
    return $rv
  }

  #
  # Iterates through $::argv and places all --key=val pairs in
  # flagsVarName and non-flag arguments in nonFlagsVarName. Flags
  # without no = in them are assumed to be boolean and get a value of
  # 1. This does not modify $::argv. Quirks:
  #
  # - All flags may have either one or two leading dashes. The
  #   resulting dict entries are devoid of leading dashes.
  #
  # - If any --flag is provided more than once, the last one wins,
  #   overwriting any previous entry.
  #
  # - If a flag in the form "--" is encountered then all arguments
  #   after it are treated as non-flags even if they start with "-".
  #
  # Returns the number of arguments processed.
  #
  # Example:
  #
  #   collect-argv flags nonFlags
  #
  # Then $flags and $nonFlags contain their respective lists.
  #
  proc collect-argv {flagsVarName nonFlagsVarName} {
    upvar $flagsVarName liF
    upvar $nonFlagsVarName liN
    set seenDoubleDash 0
    array set liF {}
    set liN {}
    set rv 0
    foreach arg $::argv {
      incr rv
      if {$seenDoubleDash} {
        lappend liN $arg
        continue
      }
      switch -regexp -- $arg {
        ^--$ {
          set seenDoubleDash 1
        }
        ^--?[^=]+=.* {
          if {[regexp -- {--?([^=]+)=(.*)} $arg -> n v]} {
            set liF($n) $v
          } else {
            error "Unexpected regex mismatch for arg: $arg"
          }
        }
        ^--?.+ {
          regexp -- {--?(.+)} $arg -> n
          set liF($n) 1
        }
        default {
          lappend liN $arg
        }
      }
    }
    return $rv
  }

  #
  # If $key refers to a --key[=value] flag, returns its value, else
  # returns $dflt. $key may be a list of alternative names, e.g. {file
  # f}. If it is, only the first match found is returned.
  #
  proc flag-get {key {dflt ""}} {
    variable argvFlags
    foreach k $key {
      if {[info exists argvFlags($k)]} {
        return $argvFlags($k)
      }
    }
    return $dflt
  }

  #
  # Works like [flag-get] but also removes the flag(s) from this
  # namespace's view of the argv flags.
  #
  proc flag-take {key {dflt ""}} {
    variable argvFlags
    set rv [flag-get $key $dflt]
    foreach k $key {
      array unset argvFlags $k
    }
    return $rv
  }

  #
  # Returns the number of flags in this namespace's view of the argv
  # flags.
  #
  proc flag-count {} {
    variable argvFlags
    return [array size argvFlags]
  }

  #
  # Returns true if the argv flag --$key[=...] was set, else false.
  # $key may be a list of alternative names, e.g. {file f}
  #
  proc flag-isset {key} {
    variable argvFlags
    foreach k $key {
      if {[info exists argvFlags($k)]} {
        return 1
      }
    }
    return 0
  }

  #
  # @file-content ?-trim? filename
  #
  # Opens the given file, reads all of its content, and returns it.  If
  # the first arg is -trim, the contents of the file named by the second
  # argument are trimmed before returning them.
  #
  proc file-content {args} {
    set trim 0
    set fname $args
    if {"-trim" eq [lindex $args 0]} {
      set trim 1
      lassign $args - fname
    }
    set fp [open $fname rb]
    set rc [read $fp]
    close $fp
    if {$trim} { return [string trim $rc] }
    return $rc
  }

  #
  # @file-conent filename
  #
  # Returns the contents of the given file as an array of lines, with
  # the EOL stripped from each input line.
  #
  proc file-content-list {fname} {
    set fp [open $fname rb]
    set rc {}
    while { [gets $fp line] >= 0 } {
      lappend rc $line
    }
    close $fp
    return $rc
  }

  #
  # @file-write ?-ro? fname content
  #
  # Works like autosetup's [writefile] but explicitly uses binary mode
  # to avoid EOL translation on Windows. If $fname already exists, it is
  # overwritten, even if it's flagged as read-only.
  #
  proc file-write {args} {
    if {"-ro" eq [lindex $args 0]} {
      lassign $args ro fname content
    } else {
      set ro ""
      lassign $args fname content
    }
    file delete -force -- $fname; # in case it's read-only
    set f [open $fname wb]
    puts -nonewline $f $content
    close $f
    if {"" ne $ro} {
      catch {
        exec chmod -w $fname
        #file attributes -w $fname; #jimtcl has no 'attributes'
      }
    }
  }

  #
  # @parse-simple-flags ...
  #
  # A helper to parse flags from proc argument lists.
  #
  # Expects a list of arguments to parse, an array name to store any
  # -flag values to, and a prototype object which declares the flags.
  #
  # The prototype must be a list in one of the following forms:
  #
  #   -flag defaultValue {script}
  #
  #   -flag => defaultValue
  #   -----^--^ (with spaces there!)
  #
  # Repeated for each flag.
  #
  # The first form represents a basic flag with no associated
  # following argument. The second form extracts its value
  # from the following argument in $argvName.
  #
  # The first argument to this function is the name of a var holding the
  # args to parse. It will be overwritten, possibly with a smaller list.
  #
  # The second argument the name of an array variable to create in the
  # caller's scope. (Pneumonic: => points to the next argument.)
  #
  # For the first form of flag, $script is run in the caller's scope if
  # $argv contains -flag, and the result of that script is the new value
  # for $tgtArrayName(-flag). This function intercepts [return $val]
  # from $script. Any empty script will result in the flag having ""
  # assigned to it.
  #
  # The args list is only inspected until the first argument which is
  # not described by $prototype. i.e. the first "non-flag" (not counting
  # values consumed for flags defined like --flag=>default).
  #
  # If a "--" flag is encountered, no more arguments are inspected as
  # flags. If "--" is the first non-flag argument, the "--" flag is
  # removed from the results but all remaining arguments are passed
  # through. If "--" appears after the first non-flag, it is retained.
  #
  # This function assumes that each flag is unique, and using a flag
  # more than once behaves in a last-one-wins fashion.
  #
  # Any argvName entries not described in $prototype are not treated as
  # flags.
  #
  # Returns the number of flags it processed in $argvName.
  #
  # Example:
  #
  # set args [list -foo -bar {blah} 8 9 10]
  # set args [parse-simple-flags args flags {
  #   -foo    0  {expr 1}
  #   -bar    => 0
  #   -no-baz 2  {return 0}
  # }
  #
  # After that $flags would contain {-foo 1 -bar {blah} -no-baz 2}
  # and $args would be {8 9 10}.
  #
  # Potential TODOs: consider using lappend instead of set so that any
  # given flag can be used more than once. Or add a syntax to indicate
  # that.
  #
  proc parse-simple-flags {argvName tgtArrayName prototype} {
    upvar $argvName argv
    upvar $tgtArrayName tgt
    array set dflt {}
    array set scripts {}
    array set consuming {}
    set n [llength $prototype]
    # Figure out what our flags are...
    for {set i 0} {$i < $n} {incr i} {
      set k [lindex $prototype $i]
      #puts "**** #$i of $n k=$k"
      assert {[string match -* $k]} \
        "Invalid flag value: $k"
      set v ""
      set s ""
      switch -exact -- [lindex $prototype [expr {$i + 1}]] {
        => {
          incr i 2
          if {$i >= $n} {
            error "Missing argument for $k => flag"
          }
          set consuming($k) 1
          set v [lindex $prototype $i]
        }
        default {
          set v [lindex $prototype [incr i]]
          set s [lindex $prototype [incr i]]
          set scripts($k) $s
        }
      }
      #puts "**** #$i of $n k=$k v=$v s=$s"
      set dflt($k) $v
    }
    # Now look for those flags in the source list
    array set tgt [array get dflt]
    unset dflt
    set rc 0
    set rv {}
    set skipMode 0
    set n [llength $argv]
    for {set i 0} {$i < $n} {incr i} {
      set arg [lindex $argv $i]
      if {$skipMode} {
        lappend rv $arg
      } elseif {"--" eq $arg} {
        incr skipMode
      } elseif {[info exists tgt($arg)]} {
        if {[info exists consuming($arg)]} {
          if {$i + 1 >= $n} {
            assert 0 {Cannot happen - bounds already checked}
          }
          set tgt($arg) [lindex $argv [incr i]]
        } elseif {"" eq $scripts($arg)} {
          set tgt($arg) ""
        } else {
          #puts "**** running scripts($arg) $scripts($arg)"
          set code [catch {uplevel 1 $scripts($arg)} xrc xopt]
          #puts "**** tgt($arg)=$scripts($arg) code=$code rc=$rc"
          if {$code in {0 2}} {
            set tgt($arg) $xrc
          } else {
            return {*}$xopt $xrc
          }
        }
        incr rc
      } else {
        incr skipMode
        lappend rv $arg
      }
    }
    set argv $rv
    return $rc
  }

  set subs {
    affirm
    assert
    collect-argv
    flag-count
    flag-get
    flag-isset
    flag-take
    file-content file-content-list file-write
    remove-arg
    unique-name
  }

  namespace ensemble create -subcommands $subs
  namespace export {*}$subs
  unset subs

  # initializer
  collect-argv argvFlags argvNonFlags
}; # namespace wh
